--- src/http/ngx_http_request.h.orig	Mon Sep 20 08:36:57 2010
+++ src/http/ngx_http_request.h	Mon Sep 20 10:44:21 2010
@@ -530,6 +530,9 @@
     u_char                           *header_start;
     u_char                           *header_end;
 
+    ngx_array_t                      *header_multiline; /* of ngx_str_t */
+    ngx_uint_t                        header_rewind;
+
     /*
      * a memory that can be reused after parsing a request line
      * via ngx_http_ephemeral_t
--- src/http/ngx_http_parse.c.orig	Mon Sep 20 07:09:02 2010
+++ src/http/ngx_http_parse.c	Mon Sep 20 20:48:38 2010
@@ -749,13 +749,32 @@
     return NGX_OK;
 }
 
+void
+ngx_http_parse_header_multiline_save_current_line(ngx_http_request_t *r)
+{
+    ngx_str_t  *line;
 
+    if (r->header_multiline == NULL) {
+        r->header_multiline = ngx_array_create(r->pool, 4, sizeof(ngx_str_t));
+        if (r->header_multiline == NULL) {
+            ngx_http_finalize_request(r, NGX_HTTP_INTERNAL_SERVER_ERROR);
+            return;
+        }
+    }
+
+    line = ngx_array_push(r->header_multiline);
+    line->len = r->header_end - r->header_start;
+    line->data = r->header_start;
+}
+
 ngx_int_t
 ngx_http_parse_header_line(ngx_http_request_t *r, ngx_buf_t *b,
     ngx_uint_t allow_underscores)
 {
-    u_char      c, ch, *p;
-    ngx_uint_t  hash, i;
+    u_char       c, ch, *p, *unfolded;
+    ngx_uint_t   hash, i;
+    ngx_str_t   *line;
+    size_t       size;
     enum {
         sw_start = 0,
         sw_name,
@@ -763,6 +782,7 @@
         sw_value,
         sw_space_after_value,
         sw_ignore_line,
+        sw_almost_done_or_multiline,
         sw_almost_done,
         sw_header_almost_done
     } state;
@@ -783,6 +803,11 @@
     hash = r->header_hash;
     i = r->lowcase_index;
 
+    if (r->header_rewind) {
+        r->header_rewind = 0;
+        b->pos--;
+    }
+
     for (p = b->pos; p < b->last; p++) {
         ch = *p;
 
@@ -883,6 +908,7 @@
         case sw_space_before_value:
             switch (ch) {
             case ' ':
+            case '\t':
                 break;
             case CR:
                 r->header_start = p;
@@ -909,11 +935,18 @@
                 break;
             case CR:
                 r->header_end = p;
-                state = sw_almost_done;
+                state = sw_almost_done_or_multiline;
                 break;
             case LF:
-                r->header_end = p;
-                goto done;
+                if (p + 1 >= b->last) {
+                    /* can't read ahead */
+                    r->header_rewind = 1;
+                    break;
+                } else if ((*(p + 1) != ' ') && (*(p + 1) != '\t')) {
+                    /* header is single-line */
+                    r->header_end = p;
+                    goto done;
+                }
             }
             break;
 
@@ -921,12 +954,25 @@
         case sw_space_after_value:
             switch (ch) {
             case ' ':
+            case '\t':
                 break;
             case CR:
-                state = sw_almost_done;
+                state = sw_almost_done_or_multiline;
                 break;
             case LF:
-                goto done;
+                if (p + 1 >= b->last) {
+                    /* can't read ahead */
+                    r->header_rewind = 1;
+                    break;
+                } else if ((*(p + 1) == ' ') || (*(p + 1) == '\t')) {
+                    /* header is multi-line */
+                    ngx_http_parse_header_multiline_save_current_line(r);
+                    state = sw_space_before_value;
+                } else {
+                    /* header is single-line */
+                    goto done;
+                }
+                break;
             default:
                 state = sw_value;
                 break;
@@ -944,6 +990,31 @@
             }
             break;
 
+        /* is it multi-line header? */
+        case sw_almost_done_or_multiline:
+            switch(ch) {
+            case LF:
+                if (p + 1 >= b->last) {
+                    /* can't read ahead */
+                    r->header_rewind = 1;
+                    break;
+                } else if ((*(p + 1) == ' ') || (*(p + 1) == '\t')) {
+                    /* header is multi-line */
+                    ngx_http_parse_header_multiline_save_current_line(r);
+                    state = sw_space_before_value;
+                } else {
+                    /* header is single-line, rewind */
+                    p--;
+                    state = sw_almost_done;
+                }
+                break;
+            default:
+                /* header is single-line, rewind */
+                p--;
+                state = sw_almost_done;
+            }
+            break;
+
         /* end of header line */
         case sw_almost_done:
             switch (ch) {
@@ -980,6 +1051,44 @@
     r->state = sw_start;
     r->header_hash = hash;
     r->lowcase_index = i;
+
+    if (r->header_multiline && r->header_multiline->nelts) {
+        /* combine all lines */
+        size = 0;
+
+        line = r->header_multiline->elts;
+        for (i = 0; i < r->header_multiline->nelts; i++) {
+            size += line[i].len;
+        }
+
+        size += r->header_end - r->header_start
+              + r->header_multiline->nelts + 1;
+
+        unfolded = ngx_pnalloc(r->pool, size);
+        if (unfolded == NULL) {
+            ngx_http_finalize_request(r, NGX_HTTP_INTERNAL_SERVER_ERROR);
+            return NGX_ERROR;
+        }
+
+        p = unfolded;
+        for (i = 0; i < r->header_multiline->nelts; i++) {
+            p = ngx_copy(p, line[i].data, line[i].len);
+            *p++ = ' ';
+        }
+
+        p = ngx_copy(p, r->header_start, r->header_end - r->header_start);
+        *p++ = '\0';
+
+        if ((size_t) (p - unfolded) != size) {
+            ngx_http_finalize_request(r, NGX_HTTP_INTERNAL_SERVER_ERROR);
+            return NGX_ERROR;
+        }
+
+        r->header_start = unfolded;
+        r->header_end = p;
+
+        r->header_multiline->nelts = 0;
+    }
 
     return NGX_OK;
 
